from django.contrib.contenttypes.models import ContentType

import pytest

from aleksis.apps.csv_import.field_types import (
    IsActiveFieldType,
    ShortNameFieldType,
    UniqueReferenceFieldType,
    field_type_registry,
)
from aleksis.apps.csv_import.models import ImportTemplate, get_allowed_content_types_query
from aleksis.core.models import Person

pytestmark = pytest.mark.django_db


def test_limit_content_types():
    query = get_allowed_content_types_query()
    cts = ContentType.objects.filter(**query)
    model_classes = []
    for ct in cts:
        model_classes.append(ct.model_class())
        assert ct.model_class() in field_type_registry.allowed_models
    for ct in field_type_registry.allowed_models:
        assert ct in model_classes


def test_import_template_str():
    template = ImportTemplate.objects.create(
        content_type=ContentType.objects.get_for_model(Person),
        name="foo",
        verbose_name="Bar",
    )
    assert str(template) == "Bar"


def test_import_template_field():
    template = ImportTemplate.objects.create(
        content_type=ContentType.objects.get_for_model(Person),
        name="foo",
        verbose_name="Bar",
    )
    field_0 = template.fields.create(field_type=UniqueReferenceFieldType.name, index=0)
    field_1 = template.fields.create(field_type=ShortNameFieldType.name, index=1)
    field_2 = template.fields.create(field_type=IsActiveFieldType.name, index=2)

    assert field_0.field_type_class == UniqueReferenceFieldType
    assert field_1.field_type_class == ShortNameFieldType
    assert field_2.field_type_class == IsActiveFieldType

    assert template.fields.count() == 3



from django import forms
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from aleksis.apps.csv_import.models import ImportTemplate
from aleksis.core.models import SchoolTerm


class CSVUploadForm(forms.Form):
    csv = forms.FileField(label=_("CSV file"))
    school_term = forms.ModelChoiceField(
        queryset=SchoolTerm.objects.all(), label=_("Related school term"),
    )
    template = forms.ModelChoiceField(
        queryset=ImportTemplate.objects.all(), label=_("Import template")
    )

    def __init__(self, *args, **kwargs):
        try:
            school_terms = SchoolTerm.objects.on_day(timezone.now().date())
            kwargs["initial"] = {
                "school_term": school_terms[0] if school_terms.exists() else None
            }
        except SchoolTerm.DoesNotExist:
            pass
        super().__init__(*args, **kwargs)



import os
from typing import Sequence

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model

import toml

from .field_types import FieldType, field_type_registry
from .models import ImportTemplate, ImportTemplateField


def update_or_create_template(
    model: Model,
    name: str,
    verbose_name: str,
    extra_args: dict,
    fields: Sequence[FieldType],
):
    """Update or create an import template in database."""
    ct = ContentType.objects.get_for_model(model)
    template, updated = ImportTemplate.objects.update_or_create(
        name=name,
        defaults={"verbose_name": verbose_name, "content_type": ct, **extra_args},
    )

    for i, field in enumerate(fields):
        ImportTemplateField.objects.update_or_create(
            template=template, index=i, defaults={"field_type": field.name},
        )

    ImportTemplateField.objects.filter(template=template, index__gt=i).delete()


def update_or_create_default_templates():
    """Update or create default import templates."""
    template_defs = toml.load(
        os.path.join(os.path.dirname(__file__), "default_templates.toml")
    )

    for name, defs in template_defs.items():
        model = apps.get_model(defs["model"])
        fields = [
            field_type_registry.get_from_name(field_type)
            for field_type in defs["fields"]
        ]

        update_or_create_template(
            model,
            name=name,
            verbose_name=defs.get("verbose_name", ""),
            extra_args=defs.get("extra_args", {}),
            fields=fields,
        )



from django.utils.translation import gettext as _

from jsonstore import CharField

from aleksis.apps.csv_import.field_types import field_type_registry

for model in field_type_registry.allowed_models:
    model.field(
        import_ref_csv=CharField(verbose_name=_("CSV import reference"), blank=True)
    )



from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _

from aleksis.apps.csv_import.models import ImportTemplate
from aleksis.apps.csv_import.util.process import import_csv
from aleksis.core.util import messages


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "csv_path", help=_("Path to CSV file with exported teachers"), required=True
        )
        parser.add_argument(
            "template",
            help=_("Name of import template which should be used"),
            required=True,
        )

    def handle(self, *args, **options):
        template_name = options["template"]
        try:
            template = ImportTemplate.objects.get(name=template_name).pk
        except ImportTemplate.DoesNotExist:
            messages.error(None, _("The provided template does not exist."))
            return

        import_csv(None, template, options["csv_path"])



from typing import Optional, Union

from django.contrib import messages
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

import pandas
from pandas.errors import ParserError

from aleksis.apps.csv_import.field_types import (
    DirectMappingFieldType,
    MultipleValuesFieldType,
    field_type_registry,
)
from aleksis.apps.csv_import.models import ImportTemplate
from aleksis.apps.csv_import.settings import FALSE_VALUES, TRUE_VALUES
from aleksis.apps.csv_import.util.import_helpers import has_is_active_field, is_active
from aleksis.core.models import Group, Person, SchoolTerm
from aleksis.core.util.core_helpers import DummyRecorder, celery_optional_progress


@celery_optional_progress
def import_csv(
    recorder: Union["ProgressRecorder", DummyRecorder],
    template: int,
    filename: str,
    school_term: Optional[int] = None,
) -> None:
    csv = open(filename, "rb")

    template = ImportTemplate.objects.get(pk=template)
    model = template.content_type.model_class()

    if school_term:
        school_term = SchoolTerm.objects.get(pk=school_term)

    data_types = {}
    cols = []
    cols_for_multiple_fields = {}
    for field in template.fields.all():
        field_type = field.field_type_class
        column_name = field_type.column_name

        # Get column header
        if issubclass(field_type, MultipleValuesFieldType):
            cols_for_multiple_fields.setdefault(field_type, [])
            cols_for_multiple_fields[field_type].append(column_name)

        # Get data type
        data_types[column_name] = field_type.data_type

        # Prepare field type for import
        field_type.prepare(school_term)

        cols.append(column_name)
        print(cols)
    try:
        data = pandas.read_csv(
            csv,
            sep=template.parsed_separator,
            names=cols,
            header=0 if template.has_header_row else None,
            dtype=data_types,
            usecols=lambda k: not k.startswith("_"),
            keep_default_na=False,
            converters=field_type_registry.converters,
            quotechar='"',
            encoding="utf-8-sig",
            true_values=TRUE_VALUES,
            false_values=FALSE_VALUES,
        )
    except ParserError as e:
        recorder.add_message(
            messages.ERROR, _(f"There was an error while parsing the CSV file:\n{e}")
        )
        return

    # Exclude all empty rows
    data = data.where(data.notnull(), None)

    all_ok = True
    inactive_refs = []
    created_count = 0

    data_as_dict = data.transpose().to_dict().values()
    recorder.total = len(data_as_dict)

    for i, row in enumerate(data_as_dict):
        # Fill the is_active field from other fields if necessary
        obj_is_active = is_active(row)
        if has_is_active_field(model):
            row["is_active"] = obj_is_active

        # Build dict with all fields that should be directly updated
        update_dict = {}
        for key, value in row.items():
            if key in field_type_registry.field_types:
                field_type = field_type_registry.get_from_name(key)
                if issubclass(field_type, DirectMappingFieldType):
                    update_dict[field_type.db_field] = value

        # Set alternatives for some fields
        for (
            field_type_origin,
            alternative_name,
        ) in field_type_registry.alternatives.items():
            if (
                model in field_type_origin.models
                and field_type_origin.name not in row
                and alternative_name in row
            ):
                update_dict[field_type_origin.name] = row[alternative_name]

        if template.group_type and model == Group:
            update_dict["group_type"] = template.group_type

        get_dict = {}
        match_field_found = False
        for (priority, match_field_type,) in field_type_registry.match_field_types:
            if match_field_type.name in row:
                get_dict[match_field_type.db_field] = row[match_field_type.name]
                match_field_found = True
                break

        if not match_field_found:
            raise ValueError(_("Missing unique reference."))

        if hasattr(model, "school_term") and school_term:
            get_dict["school_term"] = school_term

        if obj_is_active:
            created = False

            try:
                get_dict["defaults"] = update_dict

                instance, created = model.objects.update_or_create(**get_dict)

                # Get values for multiple fields
                values_for_multiple_fields = {}
                for field_type, cols_for_field_type in cols_for_multiple_fields.items():
                    values_for_multiple_fields[field_type] = []
                    for col in cols_for_field_type:
                        value = row[col]
                        values_for_multiple_fields[field_type].append(value)

                    # Process
                    field_type().process(
                        instance, values_for_multiple_fields[field_type]
                    )

                # Process field types with custom logic
                for process_field_type in field_type_registry.process_field_types:
                    if process_field_type.name in row:
                        try:
                            process_field_type().process(
                                instance, row[process_field_type.name]
                            )
                        except RuntimeError as e:
                            recorder.add_message(messages.ERROR, str(e))

                if template.group and isinstance(instance, Person):
                    instance.member_of.add(template.group)

                if created:
                    created_count += 1

            except (ValueError, ValidationError) as e:
                recorder.add_message(
                    messages.ERROR,
                    _(f"Failed to import {model._meta.verbose_name} {row}:\n{e}"),
                    fail_silently=True,
                )
                all_ok = False

            recorder.set_progress(i + 1)
        else:
            # Store import refs to deactivate later
            try:
                obj = model.objects.get(**get_dict)
                inactive_refs.append(obj.pk)
            except model.DoesNotExist:
                pass

        # Deactivate all persons that existed but are now inactive
        affected = model.objects.filter(pk__in=inactive_refs, is_active=True).update(
            is_active=False
        )

        if affected:
            recorder.add_message(
                messages.WARNING,
                _(
                    f"{affected} existing {model._meta.verbose_name_plural} were deactivated."
                ),
            )

    if created_count:
        recorder.add_message(
            messages.SUCCESS,
            _(f"{created_count} {model._meta.verbose_name_plural} were newly created."),
        )

    if all_ok:
        recorder.add_message(
            messages.SUCCESS,
            _(f"All {model._meta.verbose_name_plural} were imported successfully."),
        )
    else:
        recorder.add_message(
            messages.WARNING,
            _(f"Some {model._meta.verbose_name_plural} failed to be imported."),
        )



from aleksis.apps.csv_import.util.import_helpers import has_is_active_field, is_active, with_prefix
from aleksis.core.models import Group, GroupType, Person


def test_is_active():
    row = {"is_active": True}
    assert is_active(row)
    row = {"is_active": False}
    assert not is_active(row)
    assert is_active({})


def test_has_is_active_field():
    assert has_is_active_field(Person)
    assert not has_is_active_field(Group)
    assert not has_is_active_field(GroupType)


def test_with_prefix():
    assert with_prefix("Foo", "Bar") == "Foo Bar"
    assert with_prefix("", "Bar") == "Bar"
    assert with_prefix(None, "Bar") == "Bar"
    assert with_prefix(None, "") == ""
    assert with_prefix("", "") == ""



import codecs

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext as _

from aleksis.apps.csv_import.field_types import field_type_registry
from aleksis.core.mixins import ExtensibleModel
from aleksis.core.models import Group, GroupType


def get_allowed_content_types_query():
    """Get all allowed content types."""
    ids = []
    for model in field_type_registry.allowed_models:
        ct = ContentType.objects.get_for_model(model)
        ids.append(ct.pk)

    return {"pk__in": ids}


class ImportTemplate(ExtensibleModel):
    content_type = models.ForeignKey(
        ContentType,
        models.CASCADE,
        verbose_name=_("Content type"),
        limit_choices_to=get_allowed_content_types_query,
    )
    name = models.CharField(max_length=255, verbose_name=_("Name"), unique=True)
    verbose_name = models.CharField(max_length=255, verbose_name=_("Name"))

    has_header_row = models.BooleanField(
        default=True, verbose_name=_("Has the CSV file an own header row?")
    )
    separator = models.CharField(
        max_length=255,
        default=",",
        verbose_name=_("CSV separator"),
        help_text=_("For whitespace use \\\\s+, for tab \\\\t"),
    )

    group = models.ForeignKey(
        Group,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Base group"),
        help_text=_(
            "If imported objects are persons, they all will be members of this group after import."
        ),
    )
    group_type = models.ForeignKey(
        GroupType,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Group type"),
        help_text=_(
            "If imported objects are groups, they all will get this group type after import."
        ),
    )

    @property
    def parsed_separator(self):
        return codecs.escape_decode(bytes(self.separator, "utf-8"))[0].decode("utf-8")

    def save(self, *args, **kwargs):
        if not self.content_type.model == "person":
            self.group = None
        if not self.content_type.model == "group":
            self.group_type = None
        super().save(*args, **kwargs)

    def __str__(self):
        return self.verbose_name

    class Meta:
        ordering = ["name"]
        verbose_name = _("Import template")
        verbose_name_plural = _("Import templates")


class ImportTemplateField(ExtensibleModel):
    index = models.IntegerField(verbose_name=_("Index"))
    template = models.ForeignKey(
        ImportTemplate,
        models.CASCADE,
        verbose_name=_("Import template"),
        related_name="fields",
    )
    field_type = models.CharField(
        max_length=255,
        verbose_name=_("Field type"),
        choices=field_type_registry.choices,
    )

    @property
    def field_type_class(self):
        return field_type_registry.get_from_name(self.field_type)

    def clean(self):
        """Validate correct usage of field types."""
        model = self.template.content_type.model_class()
        if (
            self.field_type
            not in field_type_registry.allowed_field_types_for_models[model]
        ):
            raise ValidationError(
                _("You are not allowed to use this field type in this model.")
            )

    class Meta:
        ordering = ["template", "index"]
        unique_together = ["template", "index"]
        verbose_name = _("Import template field")
        verbose_name_plural = _("Import template fields")



from django.utils.translation import gettext as _

import pycountry
from dynamic_preferences.preferences import Section
from dynamic_preferences.types import ChoicePreference, ModelChoicePreference, StringPreference

from aleksis.core.models import Group, GroupType
from aleksis.core.registries import site_preferences_registry

csv_import = Section("csv_import", verbose_name=_("CSV import"))


@site_preferences_registry.register
class GroupTypeDepartments(ModelChoicePreference):
    section = csv_import
    name = "group_type_departments"
    required = False
    default = None
    model = GroupType
    verbose_name = _("Group type for department groups")
    help_text = _("If you leave it empty, no group type will be used.")


@site_preferences_registry.register
class GroupPrefixDepartments(StringPreference):
    section = csv_import
    name = "group_prefix_departments"
    default = ""
    verbose_name = _("Prefix for long names of department groups")
    help_text = _("If you leave it empty, no prefix will be added.")


@site_preferences_registry.register
class GroupGuardians(ModelChoicePreference):
    section = csv_import
    name = "group_guardians"
    required = False
    default = None
    model = Group
    verbose_name = _("Group for guardians")
    help_text = _("If you leave it empty, no group will be used.")


@site_preferences_registry.register
class DateLanguages(StringPreference):
    section = csv_import
    name = "date_languages"
    required = False
    default = ""
    verbose_name = _("Languages for date parsing")
    help_text = _("e. g. en,es,zh-Hant")


@site_preferences_registry.register
class PhoneNumberCountry(ChoicePreference):
    section = csv_import
    name = "phone_number_country"
    required = True
    default = "GB"
    choices = [(x.alpha_2, x.alpha_2) for x in pycountry.countries]
    verbose_name = _("Country for phone number parsing")



from rules import add_perm

from aleksis.core.util.predicates import has_global_perm, has_person

import_data_predicate = has_person & has_global_perm("csv_import.import_data")
add_perm("csv_import.import_data", import_data_predicate)



import os

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render
from django.urls import reverse
from django.utils.translation import gettext as _

from rules.contrib.views import permission_required

from aleksis.core.util.core_helpers import handle_uploaded_file, path_and_rename

from .forms import CSVUploadForm
from .util.process import import_csv


@permission_required("csv_import.import_data")
def csv_import(request: HttpRequest) -> HttpResponse:
    context = {}

    upload_form = CSVUploadForm()

    if request.method == "POST":
        upload_form = CSVUploadForm(request.POST, request.FILES)

        if upload_form.is_valid():
            filename = os.path.join(
                settings.MEDIA_ROOT,
                path_and_rename(None, request.FILES["csv"].name, upload_to="tmp"),
            )
            handle_uploaded_file(request.FILES["csv"], filename)

            result = import_csv(
                upload_form.cleaned_data["template"].pk,
                filename,
                school_term=upload_form.cleaned_data["school_term"].pk,
            )

            if result:
                context = {
                    "title": _("Progress: Import data from CSV"),
                    "back_url": reverse("csv_import"),
                    "progress": {
                        "task_id": result.task_id,
                        "title": _("Import objects …"),
                        "success": _("The import was done successfully."),
                        "error": _("There was a problem while importing data."),
                    },
                }
                return render(request, "core/pages/progress.html", context)

    context["upload_form"] = upload_form

    return render(request, "csv_import/csv_import.html", context)
